/*
  Copyright 1998 David Rysdam
  This file is released under the GPL.  Details can be
  found in the file COPYING accompanying this distribution.
*/

#include <stdio.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include "xltypes.h"
#include "catdoc.h"
#include <stdlib.h>
#include <unistd.h>
#include "catdoc.h"
/************************************************************************/
/* Displays  help message                                               */
/************************************************************************/
void help (void) {
    printf("Usage:\n xls2csv [-xl] [-s charset] [-d charset] [-c char] [ -q number] files\n");
}
void print_sheet(void);
void free_sheet(void);
void print_value(char *value);
int getshort(char *rec,int offset);
int getint(char *rec,int offset);
char *format_double(char *rec,int offset,int format_code);
char *format_int(int value,int format_code);
char* format_rk(char *rec);
char *gettypename(long rectype);
void parse_sst(char *sstbuf,int bufsize);
void printhexy(unsigned char rec[], long reclen);
void process_item (int rectype, int reclen, char *rec); 
void allocate(int row,int col);
char *copy_unicode_string(unsigned char **src);
char *convert8to8(char *src,int count);
char *convert16to8(char *src,int count);
void do_table(FILE *input,char *filename);
char *mark_string(char *instr);
short int *source_charset=NULL;
CHARSET target_charset;
SUBSTMAP spec_chars;
                /* Defines unicode chars which should be
                replaced by strings before UNICODE->target chatset
                mappigs are applied i.e. TeX special chars like %
                */
SUBSTMAP replacements;
char *input_buffer, *output_buffer;
#define QUOTE_NEVER 0
#define QUOTE_SPACES_ONLY 1
#define QUOTE_ALL_STRINGS 2
#define QUOTE_EVERYTHING 3

int quote_mode = QUOTE_ALL_STRINGS;
int biff_version=0;
char cell_separator = ',';
int main(int argc, char *argv[])
{
    FILE *input;
    char *filename =NULL;
    short int *tmp_charset;
    int c;
    int i;
    
#ifdef HAVE_LANGINFO
	get_locale_charset();
#endif	
	
	check_charset(&dest_csname,dest_csname); 
	while ((c=getopt(argc,argv,"ls:d:xq:c:"))!=-1) {
	switch(c)  {
	    case 'l':
		list_charsets(); exit(0);
	    case 'x': 
		unknown_as_hex = 1; break;
	    case 's':
	        check_charset(&source_csname,optarg);
		source_charset=read_charset(source_csname);
		break;
	    case 'd':
		check_charset(&dest_csname,optarg);
		break;
	    case 'q':
		{   char *errptr;
		    quote_mode = strtol(optarg,&errptr,0);
		    if ((errptr && *errptr)||quote_mode<0||quote_mode>3) {
			fprintf(stderr,
				"argument of -q should be number from 0 to 3\n");
                        exit(1);
		    }
		}    
		break;
	    case 'c':
		cell_separator = optarg[0];
		break;
	    default:
		help();
		exit(1);
	}	
    }
		/* charset conversion init*/
    input_buffer=malloc(FILE_BUFFER);
    if (strcmp(dest_csname,"utf-8")) {
	    tmp_charset=read_charset(dest_csname);
    target_charset=make_reverse_map(tmp_charset);
    free(tmp_charset);
    } else { 
    target_charset=NULL;
    } 
    spec_chars=read_substmap(stradd("ascii",SPEC_EXT));
    if (!spec_chars) {
	fprintf(stderr,"Cannod read substitution map ascii%s\n",
		SPEC_EXT);
	exit(1);
    }  
    replacements=read_substmap(stradd("ascii",REPL_EXT));
    if (!replacements) {
        fprintf(stderr,"Cannod read substitution map ascii%s\n",
		REPL_EXT);
      exit(1);
    }  
    if (optind>=argc) {
	if (isatty(fileno(stdin))) {
	    help();
	    exit(0);
	}    
	do_table(stdin,"STDIN");
        exit (0);
    }	
    for (i=optind;i<argc;i++) {
	filename = argv[i];
	input=fopen(filename,"rb");
	if (!input) {
	    perror(filename);
	    exit(1);
	}   
        do_table(input,filename);
    }	
    return 0;
}

void do_table(FILE *input,char *filename) {    
    long rectype;
	long reclen,build_year=0,build_rel=0,offset=0;
    int eof_flag=0;
    unsigned char rec[MAX_MS_RECSIZE];
    int itemsread=1;
    while (itemsread) {
	fread(rec,2,1,input);
		biff_version=getshort(rec,0);
		if ( biff_version == 0x0809 || biff_version == 0x0409 ||
				 biff_version == 0x0209 || biff_version == 0x0009 ) {
/* 			fprintf(stderr,"Version: %04x\n",biff_version); */
	    fread(rec,2,1,input);
	    reclen=getshort(rec,0);
/* 			fprintf(stderr,"Type: %04x\n",reclen); */
	    if (reclen==8 || reclen==16) {
				if (biff_version == 0x0809 ) {
					itemsread=fread(rec,4,1,input);
					build_year=getshort(rec+2,0);
					build_rel=getshort(rec,0);
/* 					fprintf(stderr,"Year: %04d %04d\n",build_year,build_rel); */
					if(build_year > 5 ) {
						itemsread=fread(rec,8,1,input);
						biff_version=8;
						offset=12;
					}
					else {
						biff_version=7;
						offset=4;
					}
				} else if (biff_version == 0x0209 ) {
					biff_version=3;
					offset=2;
				} else if (biff_version == 0x0409 ) {
					offset=2;
					biff_version=4;
				} else {
					biff_version=2;
				}
				itemsread=fread(rec,reclen-offset,1,input);
		break;
	    } else {
		fprintf(stderr,"%s: Invalid BOF record\n",filename);
	        return;
	    } 
		} else {
			itemsread=fread(rec,126,1,input);
	}
    }    
    if (feof(input)) {
	fprintf(stderr,"%s: No BOF record found\n",filename);
	exit(1);
    }    
    while(itemsread){
	char buffer[2];
	rectype = 0;
	itemsread = fread(buffer, 2, 1, input);
	rectype=getshort(buffer,0);
	if(itemsread == 0)
	    break;
	reclen=0;

	itemsread = fread(buffer, 2, 1, input);
	reclen=getshort(buffer,0);
	if (reclen && reclen <MAX_MS_RECSIZE &&reclen >0){
	    itemsread = fread(rec, 1, reclen, input);
	    rec[reclen] = '\0';
	}
	if(eof_flag) {
	    if (rectype != BOF) {
		break;
	    }    
	}
/* 		fprintf(stderr,"Rectype %04xh reclen=%d\n",rectype, reclen); */
	process_item(rectype,reclen,rec);
	if (rectype == MSEOF) {
	    eof_flag=1;
	} else {
	    eof_flag=0;	
	}  
    }
    fclose(input);
    return;
}

struct rowdescr {
    int last, end;
    unsigned char **cells;
};	

unsigned char **sst=NULL;/* Shared string table parsed into array of strings in
			    output encoding*/
int sstsize = 0; /*Number of strings in SST*/
unsigned char *sstBuffer=NULL; /*Unparsed sst to accumulate all its parts*/
int sstBytes = 0; /*Size of SST Data, already accumulated in the buffer */
int codepage=1251; /*default*/
struct rowdescr *rowptr=NULL;
unsigned char **saved_reference=NULL;
int startrow=0,lastrow=0;
int prev_rectype=0;

void process_item (int rectype, int reclen, char *rec) {
    if (rectype != CONTINUE && prev_rectype == SST) {
	/* we have accumulated  unparsed SST, and now encountered
	 * another record, which indicates that SST is ended */
	parse_sst(sstBuffer,sstBytes);
    }	
    switch (rectype) {
	case FILEPASS: {
			   fprintf(stderr,"File is encrypted\n");
			   exit(69);
			   break;
		       }
	case 0x42: {
			   char *source_csname;			   
			   if (source_charset) break;
			   codepage=getshort(rec,0);
			   fprintf(stderr,"CODEPAGE %d\n",codepage); 
			   if (codepage!=1200) {
			       char cp[9];
			       sprintf(cp,"cp%d",codepage);
			       check_charset(&source_csname,cp);
			       source_charset=read_charset(source_csname);
			   }		 
			   break;
		       }  
	case FORMAT: {
			 /*     int i;
				char *ptr;
				printf( "Format %x \"",getshort(rec,0));	 
				if (rec[2] == reclen - 3 && rec[3] != 0) {
				for (i=0,ptr=rec+3;i<rec[2];i++,ptr++) {
				fputc(*ptr,stdout);
				}
				} else {
				for (i=0,ptr=rec+5;i<rec[2];i++,ptr+=2) {
				fputc(*ptr,stdout);
				}    
				}	 
				printf ("\"\n"); */
			 break;
		     }			 
	case SST: {
		      /* Just copy SST into buffer, and wait until we get
		       * all CONTINUE records
		       */ 
		      if (sstBuffer||sst) {
			  fprintf(stderr,"Duplicated SST found\n");
			  exit(1);
		      }    
		      sstBuffer=(char*)malloc(reclen);
		      sstBytes = reclen;
		      if (sstBuffer == NULL ) {
			  perror("SSTptr alloc error! ");
			  exit(1);
		      }	  
		      memcpy(sstBuffer,rec,reclen);
		      break;
		  }	
	case CONTINUE: {
			   if (prev_rectype != SST) {
			       /*fprintf(stderr,"Unexpected CONTINUE of record 0x%X\n",prev_rectype);*/
			       return; /* to avoid changing of prev_rectype;*/
			   }    
			   sstBuffer=realloc(sstBuffer,sstBytes+reclen);
			   if (sstBuffer == NULL ) {
			       perror("SSTptr realloc error! ");
			       exit(1);
			   }	  
			   memcpy(sstBuffer+sstBytes,rec,reclen);
			   sstBytes+=reclen;
			   return;
		       }			   
	case LABEL: {
			int row,col;
			struct rowdescr *prow;
			unsigned char *src=(unsigned char *)rec+6;

			saved_reference=NULL;
			row = getshort(rec,0); 
			col = getshort(rec,2);
			/* 		fprintf(stderr,"LABEL!\n"); */
			allocate(row,col);
			prow=rowptr+row;
			prow->cells[col]=copy_unicode_string(&src);
			break;
		    }     
	case CONSTANT_STRING: {
				  struct rowdescr *prow;
				  int row = getshort(rec,0); 
				  int col = getshort(rec,2);
				  int string_no=getshort(rec,6);
				  if (!sst) {
				      fprintf(stderr,"CONSTANT_STRING before SST parsed\n");
				      exit(1);
				  }    
				  saved_reference=NULL;
				  allocate(row,col);
				  prow=rowptr+row;
				  if (string_no>=sstsize|| string_no < 0 ) {
				     fprintf(stderr,"string index out of boundary\n"); 
				      exit(1);	 
				  } else if (sst[string_no] !=NULL) {	
				      int len;
				      char *outptr;
				      len=strlen(sst[string_no]);
				      outptr=prow->cells[col]=malloc(len+2);
				      *(outptr++)='"';
				      strcpy(outptr,sst[string_no]);
				  } else {
				      prow->cells[col]=malloc(1);
				      strcpy(prow->cells[col],"");
				  }	
				  break;
			      }
	case 0x03:
	case 0x103:
	case 0x303:
	            fprintf(stderr,"Something like number found: %X\n",rectype);		      
	case NUMBER: {
			 int row,col;
			 struct rowdescr *prow;

			 saved_reference=NULL;
			 row = getshort(rec,0)-startrow; 
			 col = getshort(rec,2);
			 allocate(row,col);
			 prow=rowptr+row;
			 prow->cells[col]=strdup(format_double(rec,6,getshort(rec,4)));
			 break;
		     }
	case INTEGER_CELL: {
			       int row,col;
			       struct rowdescr *prow;

			       row = getshort(rec,0)-startrow;
			       col = getshort(rec,2);
			       allocate(row,col);
			       prow=rowptr+row;
			       prow->cells[col]=strdup(format_int(getshort(rec,7),getshort(rec,4)));		  
			       break;

			   }				  
	case RK: {
		     int row,col;
		     struct rowdescr *prow;

		     saved_reference=NULL;
		     row = getshort(rec,0)-startrow; 
		     col = getshort(rec,2);
		     allocate(row,col);
		     prow=rowptr+row;
		     prow->cells[col]=strdup(format_rk(rec+6));
		     break;
		 }
	case MULRK: {
		       int row,col,startcol,endcol,offset;
		       struct rowdescr *prow;
			 row = getshort(rec,0)-startrow; 
			 startcol = getshort(rec,2);
		        endcol = getshort(rec,reclen-2);
			  saved_reference=NULL;
			  
			for (offset=6,col=startcol;col<=endcol;offset+=6,col++) { 
			  allocate(row,col);
			  prow=rowptr+row;
		          prow->cells[col]=strdup(format_rk(rec+offset));
			  
		        }		 
			break;
		    } 
	case FORMULA: { 
			  int row,col;
			  struct rowdescr *prow;

			  saved_reference=NULL;
			  row = getshort(rec,0)-startrow; 
			  col = getshort(rec,2);
			  allocate(row,col);
			  prow=rowptr+row;
			  if (((unsigned char)rec[12]==0xFF)&&(unsigned char)rec[13]==0xFF) {
			      /* not a floating point value */
			      if (rec[6]==1) {
				  /*boolean*/
				  char buf[2]="0";
				  buf[0]+=rec[9];
				  prow->cells[col]=strdup(buf);
			      } else if (rec[6]==2) {
				  /*error*/
				  char buf[6]="ERROR";
				  prow->cells[col]=strdup(buf);
			      } else if (rec[6]==0) {
				  saved_reference=prow->cells+col;
			      }   
			  } else {
			      int format_code=getshort(rec,4);
			      prow->cells[col]=strdup(format_double(rec,6,format_code));
			  }		 
			  break;
		      }
	case STRING: {
			 unsigned char *src=(unsigned char *)rec;
			 if (!saved_reference) {
			     fprintf(stderr,"String record without preceeding string formula\n");
			     break;
			 }
			 *saved_reference=copy_unicode_string(&src);
			 break;
		     }	    
	case BOF: {
		      if (rowptr) {
			  fprintf(stderr,"BOF when current sheet is not flushed\n");
			  free_sheet();
		      }
		      break;
		  }	  
	case MSEOF: {
			if (!rowptr) break;
			print_sheet();
			free_sheet();
			break;
		    }
	case ROW: { 
		      /*  		fprintf(stderr,"Row! %d %d %d\n",getshort(rec,0), getshort(rec+2,0),getshort(rec+4,0));  */
		      break; 
		  } 
	case INDEX: {
			/* 		fprintf(stderr,"INDEX! %d %d\n", getint(rec+4,0), getint(rec+8,0));  */
			break;
		    }
	default: {
#if	0	
		     fprintf(stderr,"Unknown record 0x%x\n length %d\n",rectype,reclen);		
#endif 
		 }
    }
    prev_rectype=rectype; 
}  

char *mark_string(char *instr) {
   int len=0;
   char *out;
   strlen(instr);
   if (instr==NULL) {
       out = malloc(2);
       strcpy(out,"");
   } else {
       out = malloc(len+2);
       *out='"';
       strcpy(out+1,instr);
       free (instr);
   }    
   return out;
}    

char *copy_unicode_string (unsigned char **src) {
	int count=0;
	int flags = 0;
	int start_offset=0;
	int to_skip=0;								/*     
																 *    */
	int offset = 1;								/*        */
	int charsize;
/* 	char *realstart=*src; */
	char *dest;										/*     */
	char *s,*d,*c;
	
	int i,u,l,len;

/* 	for(i=0;i<20;i++) */
/* 		fprintf(stderr,"%02x ",(*src)[i]); */
/* 	fprintf(stderr,"\n"); */

	flags = *(*src+1+offset);
	if (! ( flags == 0 || flags == 1 || flags == 8 || flags == 9 ||
		    flags == 4 || flags == 5 || flags == 0x0c || flags == 0x0d ) ) {
	    count=**src;
	    flags = *(*src+offset);
	    offset --;
	}
	else {
	    count=getshort(*src,0);
	}   
	charsize=(flags &0x01) ? 2 : 1;

	switch (flags & 12 ) {
	case 0x0c: /* Far East with RichText formating */
		to_skip=4*getshort(*src,2+offset)+getint(*src, 4+offset);
		start_offset=2+offset+2+4;
/* 		fprintf(stderr,"Far East with RichText formating\n"); */
		break;
		
	case 0x08: /* With RichText formating */
		to_skip=4*getshort(*src,2+offset);
		start_offset=2+offset+2;
/* 		fprintf(stderr,"With RichText formating %d\n",getshort(*src,2+offset)); */
		break;

	case 0x04: /* Far East */
		to_skip=getint(*src, 2+offset);
		start_offset=2+offset+4;
/* 		fprintf(stderr,"Far East\n"); */
		break;

	default:
		to_skip=0;
		start_offset=2+offset;
/* 		fprintf(stderr,"Default string\n"); */
}

/* 	fprintf(stderr,"count=%d skip=%d start_offset=%d\n",count, to_skip, start_offset); */
/*     	*/
	if ( (dest=malloc(count+1)) == NULL ) {
		perror("Dest string alloc error");
		*src+=(to_skip+start_offset+(count*charsize));
		exit(0);
	}
	*src+=start_offset;
  len = count;
  *dest=0;l=0;
	for (s=*src,d=dest,i=0;i<count;i++,s+=charsize) {
/* 		fprintf(stderr,"l=%d len=%d count=%d charsize=%d\n",l,len,count,charsize); */
		if ( (charsize == 1 && (*s == 1 || *s == 0)) ||
				 (charsize == 2 && (*s == 1 || *s == 0) && *(s+1) != 4)) {
/* 			fprintf(stderr,"extchar (unicode)=%02x %02x\n",*s, *(s+1)); */
			charsize=(*s &0x01) ? 2 : 1;
			if (charsize == 2)
				s-=1;
			count++;
			continue;
		}
		if ( charsize == 2 ){
			u=(unsigned short)getshort(s,0);
			c=convert_char(u);
/* 			fprintf(stderr,"char=%02x %02x\n", *s, *(s+1)); */
		} else {
	      if (source_charset) {
		 u=(unsigned short)to_unicode(source_charset,(unsigned char)*s);
	      }	else {
		 u=(unsigned char)*s;
	      }	 
	      c=convert_char(u);
	}
		if (c != NULL) {
      l+=strlen(c);
		}
      while (l>len) {
	 len+=16;
	 dest=realloc(dest,len+1);
	 d=dest+l;
      }
      strcpy(d,c);
		if (c != NULL) {
      d+=strlen(c);
  }      
	}
	*src=s+to_skip;
  return dest;
}

void allocate (int row,int col) {
   int newrow,newcol;
   if (row>=lastrow) {
      newrow=(row/16+1)*16;
      rowptr=realloc(rowptr,newrow*sizeof(struct rowdescr));
      memset(rowptr+lastrow,0,(newrow-lastrow)*sizeof(struct rowdescr));
      lastrow=newrow;
   }
   if (col>=rowptr[row].end) {
      newcol=(col/16+1)*16;
      rowptr[row].cells=realloc(rowptr[row].cells,newcol*sizeof(char *));
      memset(rowptr[row].cells+rowptr[row].end,0,(newcol-rowptr[row].end)
	    *sizeof(char *));
      rowptr[row].end=newcol;
   }  
   if (col>rowptr[row].last) rowptr[row].last=col;
}
char *format_double(char *rec,int offset,int format_code) {	
     static char buffer [128];
     union { char cc[8];
             double d;} dconv;
     char *d,*s;
     int i;
# ifdef WORDS_BIGENDIAN     
     for(s=rec+offset+8,d=dconv.cc,i=0;
       i<8;i++) *(d++)=*(--s);
# else       
     for(s=rec+offset,d=dconv.cc,i=0;
       i<8;i++) *(d++)=*(s++);
# endif     
     sprintf(buffer,"%.10g",dconv.d);
     return buffer;
}

char *format_int(int value,int format_code) {
	static char buffer[12];
	sprintf(buffer,"%i",value);
	return buffer;
}	

char* format_rk(char *rec) {
	static char buffer [128];
	double value=0.0;
	int i;
	
	if ( *(rec) & 0x02 )
	{
		value=(double)(getint(rec,0)>>2);
	}
	else { 
		union { char cc[8];
			double d;} dconv;
		char *d,*s;
		for(i=0;i<8;i++)
			dconv.cc[i]='\0';
# ifdef WORDS_BIGENDIAN     
		for(s=rec+4,d=dconv.cc,i=0;
				i<4;i++) *(d++)=*(--s);
		dconv.cc[0]=dconv.cc[0] & 0xfc;
# else       
		for(s=rec,d=dconv.cc+4,i=0;
				i<4;i++) *(d++)=*(s++);
		dconv.cc[4]=dconv.cc[4] & 0xfc;
		value=dconv.d;
# endif     
	}
	if ( *(rec) & 0x01 ) 
		value=value*0.01;
	sprintf(buffer,"%.10g",value);
	return buffer;
}

int getshort(char *rec,int offset) {
    return (signed short int) (*((unsigned char *)(rec+offset))|
              ((*((unsigned char *)(rec+offset+1)))<<8));
}	      

int getint(char *rec,int offset) {
	return (signed int) (*((unsigned char *)(rec+offset)) |
											 ((*((unsigned char *)(rec+offset+1)))<<8) |
											 ((*((unsigned char *)(rec+offset+2)))<<16) |
											 ((*((unsigned char *)(rec+offset+3)))<<24));
}	      

void free_sheet(void) {
   int i,j;
   struct rowdescr *row;
   unsigned char **col;
   for (row=rowptr,i=0;i<lastrow;i++,row++) {
       if (!row->cells) continue;
       for (col=row->cells,j=0;j<row->end;j++,col++) {
           if (*col) {
	      free(*col);
	   }
       }
       free (row->cells);
   }
   free (rowptr);
   rowptr=NULL;
   lastrow=0;
}

void print_sheet(void) {
	int i,j,printed=0;
    struct rowdescr *row;
    unsigned char **col;
    lastrow--;
    while (lastrow>0&&!rowptr[lastrow].cells) lastrow--;
    for(i=0,row=rowptr;i<=lastrow;i++,row++) {
       if (row->cells) {
	  for (j=0,col=row->cells;j<=row->last;j++,col++) {
				if (j){
					fputc(cell_separator,stdout);
					printed=1;
				}
	     if (*col) {
	        print_value(*col);
					printed=1;
	     }
	  }
			if (printed) {
       fputc('\n',stdout);
				printed=0;
    }
}
	}
}

time_t float2date(double f) { 
/* Hacked version. Excell stores date as floating point count of days
 * since 1.1.1900. We are substracting value of 1.1.1970 and multiplying
 * by 86400 thus getting seconds from the epoch
 */
 return rint((f-25569.0)*86400); 
}
void print_value(char *value) 
{
  int i,len;
  int quotes=0;
  int is_string=(*value == '"');
  char *ptr = is_string?value+1:value;
	if (ptr != NULL) {
  len=strlen(ptr);
	} else {
		len = 0;
	}
  switch (quote_mode) {
      case QUOTE_NEVER:
	  break;
      case QUOTE_SPACES_ONLY:	  
	  for (i=0;i<len;i++) {
	     if (isspace(ptr[i]) || ptr[i]==cell_separator || 
		 ptr[i]=='"') {
		 quotes=1;
		 break;
             }
	  }    
	  break;
      case QUOTE_ALL_STRINGS:
	  quotes=is_string;
	  break;
      case QUOTE_EVERYTHING:
	  quotes = 1;
          break;     
   }  	  
  if (quotes) {
    fputc('\"',stdout);
    for (i=0;i<len;i++) {
      if (ptr[i]=='\"') {
         fputc('\"',stdout);
         fputc('\"',stdout);
      } else {
         fputc(ptr[i],stdout);
      }
    }   
    fputc('\"',stdout);
  } else {
    fputs(ptr,stdout);
  }
}    

void parse_sst(char *sstbuf,int bufsize) {
	int i; /* index into sst */
        char *curString; /* pointer into unparsed buffer*/
	char *barrier=sstbuf+bufsize; /*pointer to end of buffer*/
	char **parsedString;/*pointer into parsed array*/ 
	sstsize = getint(sstbuf+4,0);
	sst=malloc(sstsize*sizeof(char *));
	if (sst == NULL) {
	    perror("SST allocation error");
	    exit(1);
	}
	for (i=0,parsedString=sst,curString=sstbuf+8;
		i<sstsize && curString<barrier; i++,parsedString++) {

	       *parsedString = copy_unicode_string(&curString);
	}       
}	
